#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
#include <assert.h>

#include "xmalloc.h"
#include "buffer.h"
#include "misc.h"
#include "io.h"
#include "list.h" 
#include "hal-proto.h"
#include "hal-cmd-parser.h"
#include "hal-cmd.h"
#include "hal-peer.h"
#include "hal-srv.h"

//#define DEBUG_HAL_SRV

#ifdef DEBUG_HAL_SRV
#define debug(fmt, arg...) fprintf(stderr, "debug: %s: "fmt CRLF, \
						__FUNCTION__, ##arg)
#else
#define debug(fmt, arg...) do {} while(0)
#endif

#define warn(fmt, arg...) fprintf(stderr, "warn: %s: "fmt CRLF, \
						__FUNCTION__,  ##arg)

#define info(fmt, arg...) fprintf(stderr, "warn: %s: "fmt CRLF, \
						__FUNCTION__,  ##arg)

/* Structure represents a single connection to server. */
struct hal_srv {
	struct hal_peer_base	base;
	int			connected:1;
};

static struct slist		*srv_list;
static struct io_context	*io;
static struct hal_srv_callbacks	callbacks;

/* Handler for the connect request. A user provides us with
 * inputs, outputs image sizes and the role type (slave or master). */
static void connect_request_cb(void *opaque)
{
	struct hal_srv *srv;
	const char *status;
	unsigned int inputs_sz, outputs_sz;
	int master;

	srv = (struct hal_srv *)opaque;

	debug("CN request");

	if (srv->connected) {
		hal_status_error_reply(HAL_PEER_BASE(srv)->wb, HAL_CMD_CONNECT,
					       "Peer has already connected");
		goto send;
	}

	assert(callbacks.connect_callback != NULL);

	status = callbacks.connect_callback(&inputs_sz, &outputs_sz, &master);
	if (status == HAL_OK) {
		hal_cmd_connect_reply(HAL_PEER_BASE(srv)->wb, inputs_sz, outputs_sz, master);
		srv->connected = 1;
	} else {
		hal_status_error_reply(HAL_PEER_BASE(srv)->wb, HAL_CMD_CONNECT, status);
	}

	debug("CN request: %s", status);

send:
	hal_peer_send_default(HAL_PEER_BASE(srv));
}

/* Handler for several request commands. Read/Write end and Read/Write prepare
 * commands are common. */
static void common_request(const char *cmd, void *opaque)
{
	struct hal_srv *srv;
	const char *(*user_cb)(void);
	void (*cmd_reply)(struct buffer *);
	const char *status;

	srv = (struct hal_srv *)opaque;
	user_cb = NULL;

	debug("%s request", cmd);

	if (!srv->connected) {
		hal_status_error_reply(HAL_PEER_BASE(srv)->wb, cmd,
						"Peer is not connected");
		goto send;
	}

	if (cmd == HAL_CMD_READ_PREPARE) {
		user_cb = callbacks.read_prepare_callback;
		cmd_reply = hal_cmd_read_prepare_reply;
	} else if (cmd == HAL_CMD_READ_END) {
		user_cb = callbacks.read_end_callback;
		cmd_reply = hal_cmd_read_end_reply;
	} else if (cmd == HAL_CMD_WRITE_PREPARE) {
		user_cb = callbacks.write_prepare_callback;
		cmd_reply = hal_cmd_write_prepare_reply;
	} else if (cmd == HAL_CMD_WRITE_END) {
		user_cb = callbacks.write_end_callback;
		cmd_reply = hal_cmd_write_end_reply;
	}

	assert(user_cb != NULL);

	status = user_cb();
	if (status == HAL_OK)
		cmd_reply(HAL_PEER_BASE(srv)->wb);
	else 
		hal_status_error_reply(HAL_PEER_BASE(srv)->wb, cmd, status);

	debug("%s request: %s", cmd, status);

send:
	hal_peer_send_default(HAL_PEER_BASE(srv));
}

static void read_prepare_request_cb(void *opaque)
{
	common_request(HAL_CMD_READ_PREPARE, opaque);
}

static void read_end_request_cb(void *opaque)
{
	common_request(HAL_CMD_READ_END, opaque);
}

static void write_prepare_request_cb(void *opaque)
{
	common_request(HAL_CMD_WRITE_PREPARE, opaque);
}

static void write_end_request_cb(void *opaque)
{
	common_request(HAL_CMD_WRITE_END, opaque);
}

/* Handler for write inputs and outputs request commands. */
static void write_image_request(const char *cmd, const char image[HAL_IMAGE_MAX+1],
				unsigned int off, void *opaque)
{
	struct hal_srv *srv;
	const char *(*user_cb)(const char *, unsigned int);
	void (*write_reply)(struct buffer *);
	const char *status;

	srv = (struct hal_srv *)opaque;
	user_cb = NULL;

	debug("%s request", cmd);

	if (!srv->connected) {	
		hal_status_error_reply(HAL_PEER_BASE(srv)->wb, cmd,
						"Peer is not connected");
		goto send;
	}

	if (cmd == HAL_CMD_WRITE_INPUTS) {
		user_cb = callbacks.write_inputs_callback;
		write_reply = hal_cmd_write_inputs_reply;
	} else if (cmd == HAL_CMD_WRITE_OUTPUTS) {
		user_cb = callbacks.write_outputs_callback;
		write_reply = hal_cmd_write_outputs_reply;
	}

	assert(user_cb != NULL);

	status = user_cb(image, off);
	if (status == HAL_OK)
		write_reply(HAL_PEER_BASE(srv)->wb);
	else
		hal_status_error_reply(HAL_PEER_BASE(srv)->wb, cmd, status);

	debug("%s request: status %s", cmd, status);
send:
	hal_peer_send_default(HAL_PEER_BASE(srv));
}

static void write_inputs_request_cb(const char inputs[HAL_IMAGE_MAX+1],
				    unsigned int off, void *opaque)
{
	write_image_request(HAL_CMD_WRITE_INPUTS, inputs, off, opaque);
}

static void write_outputs_request_cb(const char outputs[HAL_IMAGE_MAX+1],
				     unsigned int off, void *opaque)
{
	write_image_request(HAL_CMD_WRITE_OUTPUTS, outputs, off, opaque);
}

/* Handler for read image request commands. */
static void read_image_request(const char *cmd, unsigned int len,
			       unsigned int off, void *opaque)
{
	struct hal_srv *srv;
	const char *(*user_cb)(char *, unsigned int, unsigned int);
	void (*read_reply)(struct buffer *,  unsigned int, const char *);
	char image[HAL_IMAGE_MAX+1];
	const char *status;

	srv = (struct hal_srv *)opaque;
	user_cb = NULL;

	debug("%s request", cmd);

	if (!srv->connected) {
		hal_status_error_reply(HAL_PEER_BASE(srv)->wb, cmd,
						"Peer is not connected");
		goto send;
	}
	/* Get the right read callback. */
	if (cmd == HAL_CMD_READ_INPUTS) {
		user_cb = callbacks.read_inputs_callback;
		read_reply = hal_cmd_read_inputs_reply;
	} else if (cmd == HAL_CMD_READ_OUTPUTS) {
		user_cb = callbacks.read_outputs_callback;
		read_reply = hal_cmd_read_outputs_reply;
	}

	assert(user_cb != NULL);

	status = user_cb(image, len, off);
	if (status == HAL_OK)
		read_reply(HAL_PEER_BASE(srv)->wb, off, image);
	else
		hal_status_error_reply(HAL_PEER_BASE(srv)->wb, cmd, status);
	
	debug("%s request: status %s", cmd, status);
send:
	hal_peer_send_default(HAL_PEER_BASE(srv));
}

static void read_inputs_request_cb(unsigned int len, unsigned int off,
				   void *opaque)
{
	read_image_request(HAL_CMD_READ_INPUTS, len, off, opaque);
}

static void read_outputs_request_cb(unsigned int len, unsigned int off,
				    void *opaque)
{
	read_image_request(HAL_CMD_WRITE_INPUTS, len, off, opaque);
}

static void not_handled_cb()
{
	warn("Get unexpected command");
}

static void parser_init()
{
	struct hal_cmd_callbacks parser_callbacks;

	memset(&parser_callbacks, 0, sizeof(struct hal_cmd_callbacks));

	parser_callbacks.connect_request_callback = connect_request_cb;
	parser_callbacks.read_prepare_request_callback = read_prepare_request_cb;
	parser_callbacks.read_end_request_callback = read_end_request_cb;
	parser_callbacks.write_prepare_request_callback = write_prepare_request_cb;
	parser_callbacks.write_end_request_callback = write_end_request_cb;
	parser_callbacks.write_inputs_request_callback = write_inputs_request_cb;
	parser_callbacks.write_outputs_request_callback = write_outputs_request_cb;
	parser_callbacks.read_inputs_request_callback = read_inputs_request_cb;
	parser_callbacks.read_outputs_request_callback = read_outputs_request_cb;
	parser_callbacks.not_handled_callback = not_handled_cb;

	hal_cmd_parser_register_callbacks(&parser_callbacks);
}

static void peer_event_handler(struct hal_peer_base *peer,
			       hal_peer_event_t event)
{
	struct slist *list;

	switch (event) {
	case HAL_PEER_EVENT_ERROR:
		warn("got error event: peer is closed");
		list = slist_find(srv_list, peer, NULL);
		assert(list != NULL);
		srv_list = slist_unlink(srv_list, list);
		hal_peer_base_deinit(peer);
		xfree(peer);
		slist_free(&list);
		break;
	default:
		break;
	}
}

void hal_srv_client_new(int wfd, int rfd)
{
	struct hal_srv *srv;
	struct slist *list;

	return_if_fail(wfd >= 0);
	return_if_fail(rfd >= 0);

	srv = xmalloc0(sizeof(*srv));
	hal_peer_base_init(HAL_PEER_BASE(srv), io, wfd, rfd, peer_event_handler);

	list = slist_new();
	list->data = srv;
	srv_list = slist_add(srv_list, list);
}

int hal_srv_register_callbacks(struct hal_srv_callbacks *new_cb)
{
	return_val_if_fail(new_cb != NULL, -1);

	callbacks = *new_cb;

	return 0;
}

int hal_srv_init(struct io_context *io_ptr)
{
	return_val_if_fail(io_ptr != NULL, -1);

	io = io_ptr;
	parser_init();

	return 0;
}

void hal_srv_deinit()
{
	struct slist *iter, *next;
	struct hal_peer_base *peer;

	iter = srv_list;
	while (iter != NULL) {
		peer = (struct hal_peer_base *)iter->data;
		hal_peer_base_deinit(peer);
		xfree(peer);

		next = iter->next;
		slist_free(&iter);
		iter = next;
	}

	srv_list = NULL;
}

